#include <ESP8266WiFi.h>
#include <ArduinoOTA.h>
#include <ESP8266WebServer.h>
#include <PubSubClient.h>
#include <AccelStepper.h>
#include <Adafruit_NeoPixel.h>

#ifndef STASSID
#define STASSID ""
#define STAPSK  ""
#endif

#define MotorStatus D2 // Steppermotor aktivieren/deaktivieren
#define LEDpin 3 // Umgesteuerter GPIO für die NEO-Led
#define NUMPIXELS 1 // Anzal der LEDs

const char* ssid = STASSID;
const char* password = STAPSK;

char* mqtt_client_id = "Huehnerklappe";
const char* mqtt_server = "192.168.178.--";
const int mqtt_port = 1883;
const char* mqtt_user = "";
const char* mqtt_password = "";

const byte pinTuerAuf = D6;    // Endschalter Tür offen
const byte pinTuerZu = D5;   // Endschalter Tür geschlossen
const byte pinTasterAuf = D0; // Taster Tür öffnen
const byte pinTasterZu = D7;  // Taster Tür schließen
const byte pinTasterStopp = D1; // Taster Bewegung anhalten

const long SchritteUmdrehung = 200; // Schritte des Schrittmotors für 360 Grad der Tür
const unsigned int TuerbewegungZu =  61000;
const unsigned int TuerbewegungAuf = -61000;
enum ZUSTAENDE {STOP, TUERZU, OEFFNEN, TUERAUF, SCHLIESSEN};

int lastEndlageAufState = 1;
int endlageAufState = 1;
int lastEndlageZuState = 1 ;
int endlageZuState = 1;

int lastTasterAufState = 1;
int tasterAufState = 1;
int lastTasterZuState = 1;
int tasterZuState = 1;
int lastTasterStoppState = 1;
int tasterStoppState = 1;

int richtung = 0; //  1 > Auf, 2 > Zu 
int stopzeitTuer = 10000; //Pausenzeit bis Weiterfahrt in ms


WiFiClient espClient;
PubSubClient mqttClient(espClient);

ESP8266WebServer webserver(80);

AccelStepper stepper(1, D3, D4);  //Pins für den Steppermotor STEP , DIR
AccelStepper stepperauf(1, D3, D4);  //Pins für den Steppermotor STEP , DIR
Adafruit_NeoPixel pixels = Adafruit_NeoPixel(1, LEDpin, NEO_GRB + NEO_KHZ800);

void setup() {

  //GPIO Umsteuerung
  //GPIO 1 (TX) swap the pin to a GPIO.
  pinMode(1, FUNCTION_3);
  //GPIO 3 (RX) swap the pin to a GPIO.
  pinMode(3, FUNCTION_3);


  Serial.begin(9600);
  // Per WLAN mit dem Netzwerk verbinden
  Serial.print("Verbinden mit ");
  Serial.println(ssid);
  WiFi.begin(ssid, password);
  while (WiFi.status() != WL_CONNECTED) {
    delay(500);
    Serial.print(".");
  }

  ///////////////////////////////////////////
  // Anfang Update über Wlan
  ///////////////////////////////////////////

  ArduinoOTA.onStart([]() {
    String type;
    if (ArduinoOTA.getCommand() == U_FLASH) {
      type = "sketch";
    } else { // U_FS
      type = "filesystem";
    }

    // NOTE: if updating FS this would be the place to unmount FS using FS.end()
    Serial.println("Start updating " + type);
  });
  ArduinoOTA.onEnd([]() {
    Serial.println("\nEnd");
  });
  ArduinoOTA.onProgress([](unsigned int progress, unsigned int total) {
    Serial.printf("Progress: %u%%\r", (progress / (total / 100)));
  });
  ArduinoOTA.onError([](ota_error_t error) {
    Serial.printf("Error[%u]: ", error);
    if (error == OTA_AUTH_ERROR) {
      Serial.println("Auth Failed");
    } else if (error == OTA_BEGIN_ERROR) {
      Serial.println("Begin Failed");
    } else if (error == OTA_CONNECT_ERROR) {
      Serial.println("Connect Failed");
    } else if (error == OTA_RECEIVE_ERROR) {
      Serial.println("Receive Failed");
    } else if (error == OTA_END_ERROR) {
      Serial.println("End Failed");
    }
  });
  ArduinoOTA.begin();

  ///////////////////////////////////////////
  // Ende Update über Wlan
  ///////////////////////////////////////////

  // Die IP vom Webserver auf dem seriellen Monitor ausgeben
  Serial.println("");
  Serial.println("WLAN verbunden.");
  Serial.println("IP Adresse: ");
  Serial.println(WiFi.localIP());

  // MQTT Brocker
  // Mit ioBroker Mqtt verbinden
  mqttClient.setServer(mqtt_server, mqtt_port);//MQTT Server, - Port
  mqttClient.setCallback(callback);

  ///////////////////////////////////////////////////////

  // Webserver Seiten
  webserver.onNotFound([]() {
    webserver.send(404, "text/plain", "Link wurde nicht gefunden!");
  });
  webserver.on("/", []() {
    webserver.send(200, "text/plain", "Startseite");
  });
  webserver.on("/WebTuerAuf", []() {
    webserver.send(200, "text/plain", "Tuer wird geoeffnet.");
    WebTuerAuf();
  });
  webserver.on("/WebTuerZu", []() {
    webserver.send(200, "text/plain", "Tuer wird geschlossen.");
    WebTuerZu();
  });
  webserver.begin();
  Serial.println("Webserver gestartet.");

  //////////////////////////////////////////////////////


  pixels.begin(); // This initializes the NeoPixel library.

  Serial.println("Programmanfang");
  pinMode(LEDpin, OUTPUT); // NEO-LED
  pinMode(pinTuerAuf, INPUT_PULLUP);    // Endschalter Tür offen
  pinMode(pinTuerZu, INPUT_PULLUP);   // Endschalter Tür geschlossen
  pinMode(pinTasterAuf, INPUT_PULLUP);  // Taster Tür öffnen
  pinMode(pinTasterZu, INPUT_PULLUP); // Taster Tür schließen
  pinMode(pinTasterStopp, INPUT_PULLUP);  // Taster Bewegung anhalten
  //  pinMode(ProgrammstartTuerAuf, OUTPUT);
  stepper.setMaxSpeed(1000);
  stepperauf.setMaxSpeed(6000);
  stepper.setAcceleration(1500);
  stepperauf.setAcceleration(6000);
  pinMode(MotorStatus, OUTPUT); // Enable Motor 1
}

//ende Setup
//////////////////////////////////////////////////////////////////////////////////

void loop() {

  ArduinoOTA.handle();
  ArduinoOTA.setHostname("Huehnerklappe");

  webserver.handleClient();

  stepper.run();
  stepperauf.run();

  // Taster STOP bzw Lichtschranke
  tasterStoppState = digitalRead(pinTasterStopp);

  // Abfrage damit nur bei Änderung am Taster Aktion ausgeführt wird
  if (tasterStoppState != lastTasterStoppState) {
    if (tasterStoppState == LOW) {
      TuerSteuerung(STOP);
    }
  }

  lastTasterStoppState = tasterStoppState;

  // Taster AUF
  tasterAufState = digitalRead(pinTasterAuf);

  // Abfrage damit nur bei Änderung am Taster Aktion ausgeführt wird
  if (tasterAufState != lastTasterAufState) {
    if (tasterAufState == LOW) {
      TuerSteuerung(OEFFNEN);
    }
  }

  lastTasterAufState = tasterAufState;

  // Taster ZU
  tasterZuState = digitalRead(pinTasterZu);

  // Abfrage damit nur bei Änderung am Taster Aktion ausgeführt wird
  if (tasterZuState != lastTasterZuState) {
    if (tasterZuState == LOW) {
      TuerSteuerung(SCHLIESSEN);
    }
  }

  lastTasterZuState = tasterZuState;

  // Endlage AUF
  endlageAufState = digitalRead(pinTuerAuf);

  // Abfrage damit nicht bei jedem Durchlauf die Endlage neu gesetzt wird,
  // sondern nur bei Änderung
  if (endlageAufState != lastEndlageAufState) {
    if (endlageAufState == LOW) {
      TuerEndlagen(TUERAUF);
    }
  }

  lastEndlageAufState = endlageAufState;

  // Endlage ZU
  endlageZuState = digitalRead(pinTuerZu);

  // Abfrage damit nicht bei jedem Durchlauf die Endlage neu gesetzt wird,
  // sondern nur bei Änderung
  if (endlageZuState != lastEndlageZuState) {
    if (endlageZuState == LOW) {
      TuerEndlagen(TUERZU);
    }
  }

  lastEndlageZuState = endlageZuState;

  // MQTT Broker
  if (!mqttClient.connected()) {
    reconnectToMQTT();
  }
  mqttClient.loop();


}

//ende loop
/////////////////////////////////////////////////////////////////////////////////


long pos(int winkel) {
  return (winkel * SchritteUmdrehung / 360L);
}



// Endlagen Steuerung
void TuerEndlagen(byte zustand) {

  stopp();
  Serial.println("Motor HAlT Endlage erreicht");

  switch (zustand) {

    // Tür ist offen

    case TUERAUF:
      richtung = 0;
      Serial.println("Tuer geoeffnet");
      pixels.setPixelColor(0, pixels.Color(0, 255, 0)); //LED Farbe grün
      pixels.show();
      // Schaltzustände übermitteln
      mqttClient.publish("Huehnertuer/Innen/Tuer_oeffnet", "0");
      mqttClient.publish("Huehnertuer/Innen/Tuer_auf", "1");
      mqttClient.publish("Huehnertuer/Innen/Tuer_schliesst", "0");
      mqttClient.publish("Huehnertuer/Innen/Tuer_zu", "0");

      break;

    // Tür ist geschlossen

    case TUERZU:
      richtung = 0;
      Serial.println("Tuer geschlossen");
      pixels.setPixelColor(0, pixels.Color(255, 0, 0)); //LED Farbe rot
      pixels.show();
      // Schaltzustände übermitteln
      mqttClient.publish("Huehnertuer/Innen/Tuer_oeffnet", "0");
      mqttClient.publish("Huehnertuer/Innen/Tuer_auf", "0");
      mqttClient.publish("Huehnertuer/Innen/Tuer_schliesst", "0");
      mqttClient.publish("Huehnertuer/Innen/Tuer_zu", "1");
      break;

  }
}


/// Tuer / Motorsteuerung
void TuerSteuerung(byte zustand) {

  stopp();
  Serial.println("Motor HAlT neuer Fahrbefehl");

  digitalWrite(MotorStatus, LOW);  //Motor aktiviert

  switch (zustand) {

    // Manuelle Steuerung

    case STOP:
      stopp();
      Serial.println("Motor HAlT");
      pixels.setPixelColor(0, pixels.Color(255, 255, 255)); //LED Farbe weiss
      pixels.show();
      delay(stopzeitTuer);
      if (richtung = 1) {
        TuerSteuerung(OEFFNEN);
      }
      if (richtung = 2){
        TuerSteuerung(SCHLIESSEN); 
      }
      break;

    // Tür wird geöffnet

    case OEFFNEN:
      if ((digitalRead(pinTuerAuf)) && (digitalRead(pinTasterStopp))) {
        richtung = 1;
        stepperauf.move(pos(TuerbewegungAuf));
        Serial.print("Tuer oeffnet ... ");
        pixels.setPixelColor(0, pixels.Color(0, 0, 255)); //LED Farbe blau
        pixels.show();
        // Schaltzustände übermitteln
        mqttClient.publish("Huehnertuer/Innen/Tuer_oeffnet", "1");
        mqttClient.publish("Huehnertuer/Innen/Tuer_auf", "0");
        mqttClient.publish("Huehnertuer/Innen/Tuer_schliesst", "0");
        mqttClient.publish("Huehnertuer/Innen/Tuer_zu", "0");
      }
      break;


    // Tür wir geschlossen

    case SCHLIESSEN:
      if ((digitalRead(pinTuerZu)) && (digitalRead(pinTasterStopp))) {
        richtung = 2;
        stepper.move(pos(TuerbewegungZu));
        Serial.print("Tuer schliesst ... ");
        pixels.setPixelColor(0, pixels.Color(0, 0, 255)); //LED Farbe blau
        pixels.show();
        // Schaltzustände übermitteln
        mqttClient.publish("Huehnertuer/Innen/Tuer_oeffnet", "0");
        mqttClient.publish("Huehnertuer/Innen/Tuer_auf", "0");
        mqttClient.publish("Huehnertuer/Innen/Tuer_schliesst", "1");
        mqttClient.publish("Huehnertuer/Innen/Tuer_zu", "0");
      }
      break;
  }
}

void WebTuerAuf() {
  TuerSteuerung(OEFFNEN);
  Serial.println("Tuer oeffnet über WEB");
}
void WebTuerZu() {
  TuerSteuerung(SCHLIESSEN);
  Serial.println("Tuer schliesst über WEB");
}


void stopp() {
  stepper.stop();
  digitalWrite(MotorStatus, HIGH); //Motor deaktiviert
}



// MQTT Funktion
// *************
void reconnectToMQTT() {
  // mqttClient.setServer(mqtt_server, mqtt_port);//MQTT Server, - Port
  // mqttClient.setCallback(callback);

  while (!mqttClient.connected()) {
    if (mqttClient.connect(mqtt_client_id , mqtt_user, mqtt_password)) {
      Serial.println("Per MQTT mit ioBroker verbunden");

      mqttClient.publish("Huehnertuer/Innen/Tuer_oeffnet", "0");
      mqttClient.publish("Huehnertuer/Innen/Tuer_auf", "0");
      mqttClient.publish("Huehnertuer/Innen/Tuer_schliesst", "0");
      mqttClient.publish("Huehnertuer/Innen/Tuer_zu", "0");

      mqttClient.subscribe("Huehnertuer/Innen/Tuer");
    } else {
      Serial.print("failed with state ");
      Serial.print(mqttClient.state());
      Serial.println(" try again in 5 seconds");
      // Wait 5 seconds before retrying
      delay(5000);
    }
  }
}


// MQTT Callback
void callback(char* topic, byte* payload, unsigned int length)
{

  payload[length] = '\0';
  String strTopic = String(topic);
  String strPayload = String((char * ) payload);

  Serial.print("Message arrived [");
  Serial.print(strTopic);
  Serial.print("----");
  Serial.print(strPayload);
  Serial.print("] ");
  Serial.println();


  if (strTopic == "Huehnertuer/Innen/Tuer")
  {
    // Tuer schliessen
    if (strPayload == "false") {
      Serial.println("Tuer schliesst über IoBroker");
      TuerSteuerung(SCHLIESSEN);
    }

    // Tuer öffnen
    if (strPayload == "true") {
      Serial.println("Tuer oeffnet über IoBroker");
      TuerSteuerung(OEFFNEN);
    }
  }
}
